/*
 * palindrome.c:
 *
 * Copyright (c) 2002 James McKenzie <james@fishsoup.dhs.org>,
 * All rights reserved.
 *
 */

static char rcsid[] =
    "$Id: palindrome.c,v 1.12 2002/08/09 13:55:41 root Exp root $";

/*
 * $Log: palindrome.c,v $
 * Revision 1.12  2002/08/09 13:55:41  root
 * #
 *
 * Revision 1.11  2002/08/09 13:46:23  root
 * #
 *
 * Revision 1.10  2002/08/09 13:46:20  root
 * #
 *
 * Revision 1.9  2002/08/09 13:42:03  root
 * #
 *
 * Revision 1.8  2002/08/09 13:41:49  root
 * *** empty log message ***
 *
 * Revision 1.7  2002/08/09 13:40:44  root
 * *** empty log message ***
 *
 * Revision 1.6  2002/08/09 13:40:23  root
 * *** empty log message ***
 *
 * Revision 1.5  2002/08/06 14:07:26  root
 * *** empty log message ***
 *
 * Revision 1.4  2002/08/06 13:51:23  root
 * *** empty log message ***
 *
 * Revision 1.3  2002/08/06 11:56:07  root
 * *** empty log message ***
 *
 * Revision 1.2  2002/08/06 11:55:32  root
 * *** empty log message ***
 *
 */

#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <strings.h>
#include <stdlib.h>
#include <time.h>


#ifdef RANDOM
#define LEVELS 8
#define MINLEV 7
#else
#define LEVELS 5
#define MINLEV 0
#endif



char **words;
int nwords;
char **sdrow;
int nsdrow;


void read_words(char ***w, int *n, char *fn)
{
  int s = 1;
  char buf[1024], *ptr;
  char **wptr;
  FILE *f = fopen(fn, "r");

  *w = (char **) malloc(s * sizeof(char *));

  while (!feof(f)) {
    *buf = 0;
    fgets(buf, sizeof(buf), f);
    if (!*buf)
      break;

    ptr = index(buf, ' ');
    if (ptr)
      *ptr = 0;
    ptr = index(buf, '\n');
    if (ptr)
      *ptr = 0;
    ptr = index(buf, '\r');
    if (ptr)
      *ptr = 0;

    ptr = buf;

    (*w)[*n] = strdup(buf);
    (*n)++;

    if ((*n) == s) {
      s <<= 1;
      *w = (char **) realloc(*w, sizeof(char *) * s);
      (*w)[*n] = NULL;
    }
  }
}

char **random_lookup(char **d, int ds, char *w, int n)
{
  int l = 0, h = ds - 1;
  int lm = (l + h) / 2;
  int hm = (l + h) / 2;
  int m;


  ds--;

  while (1) {
    int c = strncasecmp(d[lm], w, n);

    if (c < 0)
      l = lm;
    if (c >= 0)
      h = lm;


    if ((l + 1) == lm) {
      break;
    }
    if (lm == 0)
      break;
    if (lm == ds)
      break;

    lm = (l + h + 1) / 2;

  }
  h = ds;
  l = 0;

  while (1) {
    int c = strncasecmp(d[hm], w, n);

    if (c <= 0)
      l = hm;
    if (c > 0)
      h = hm;


    if ((h - 1) == hm) {
      break;
    }
    if (hm == 0)
      break;
    if (hm == ds)
      break;

    hm = (l + h) / 2;

  }


  if (hm < lm)
    return NULL;

  if (hm != lm) {
    hm -= lm;
    hm++;
    m = lrand48() % hm;
    m += lm;
  } else {
    m = lm;
  }

  if (strncasecmp(d[m], w, n))
    return NULL;

  return d + m;
}



char **lookup(char **d, int ds, char *w, int n)
{
  int l = 0, h = ds - 1;
  int m = (l + h) / 2;

  ds--;

  while (1) {
    int c = strncasecmp(d[m], w, n);

    if (c < 0)
      l = m;
    if (c >= 0)
      h = m;


    if ((l + 1) == m) {
      break;
    }
    if (m == 0)
      break;
    if (m == ds)
      break;

    m = (l + h + 1) / 2;

  }

  if (strncasecmp(d[m], w, n))
    return NULL;

  return d + m;
}


char **next_match(char **m, char *w)
{
  int n = strlen(w);
  m++;

  if (!*m)
    return NULL;
  if (strncasecmp(*m, w, n))
    return NULL;

  return m;
}

report_palindrome(char *l, char *r, int level)
{
  int n = 0;
  char *buf, *ptr;

  buf = malloc(strlen(l) + strlen(r) + 4);

  if (level < MINLEV)
    return;

  buf[0] = (*l) & (255 ^ 32);
  buf[1] = 0;
  strcat(buf, l + 1);

  ptr = buf + strlen(buf);

  if (*r) {
    *(ptr++) = ' ';

    while (*(r++))
      n++;
    --r;
    --n;

    while (n--)
      *(ptr++) = (*(--r));

  }

  *(ptr++) = '.';
  *(ptr++) = '\n';
  *ptr = 0;

  fwrite(buf, 1, strlen(buf), stdout);
  free(buf);
#ifdef RANDOM
  exit(0);
#endif
}

int is_palendromic(char *u)
{
  int l = strlen(u);
  char *v = u + (l - 1);

  while (l--) {
    if (*(v--) != *(u++))
      return 0;
  }
  return 1;
}


void
palindrome_worker(char *lmatch, char *rmatch, char *unmatch, int side,
                  int level)
{
  char **wptr;
  char *next;
  int ll = strlen(lmatch);
  int rl = strlen(rmatch);
  int n;


  if (!unmatch) {
    //No slop left over - we have a palindrome
    report_palindrome(lmatch, rmatch, level);
  } else {
    //slop is palindromic - we have a palindrome
    if (is_palendromic(unmatch))
      report_palindrome(lmatch, rmatch, level);
  }

  //Give up
  if (level > LEVELS)
    return;

  //Pass 0 add any word on either side - we choose the left wlog.
  if (!unmatch) {
    int i = nwords;
    char **ptr = words;
    char *next;
    while (i--) {
      int l = strlen(*ptr);
      next = malloc(ll + l + 2);
      strcpy(next, lmatch);
      strcat(next, " ");
      strcat(next, *ptr);
      palindrome_worker(next, rmatch, *ptr, 1, level + 1);
      free(next);
      ptr++;
    }
    return;
  }

  n = strlen(unmatch);

  //Pass 1 add a whole word matching the slop and carry on on this side.
  {
    int i;
    for (i = 1; i < n; ++i) {

      if (side)
        wptr = lookup(sdrow, nsdrow, unmatch, i);
      else
        wptr = lookup(words, nwords, unmatch, i);

      if (wptr && (strlen(*wptr) == i)) {
        if (side) {
          next = malloc(rl + i + 2);
          strcpy(next, rmatch);
          strcat(next, " ");
          strcat(next, *wptr);
          palindrome_worker(lmatch, next, unmatch + i, 1, level + 1);
          free(next);
        } else {
          next = malloc(ll + i + 2);
          strcpy(next, lmatch);
          strcat(next, " ");
          strcat(next, *wptr);
          palindrome_worker(next, rmatch, unmatch + i, 0, level + 1);
          free(next);
        }
      }
    }
  }


  //Pass 2 add a part of a word matching the slop and carry on on the other side

#ifdef RANDOM
  if (side)
    wptr = random_lookup(sdrow, nsdrow, unmatch, n);
  else
    wptr = random_lookup(words, nwords, unmatch, n);

#else
  if (side)
    wptr = lookup(sdrow, nsdrow, unmatch, n);
  else
    wptr = lookup(words, nwords, unmatch, n);
#endif

  while (wptr) {
    int wl = strlen(*wptr);

    if (side) {
      next = malloc(rl + strlen(*wptr) + 2);
      strcpy(next, rmatch);
      strcat(next, " ");
      strcat(next, *wptr);

      palindrome_worker(lmatch, next, (wl == n) ? NULL : (*wptr) + n, 0,
                        level + 1);
      free(next);
    } else {
      next = malloc(ll + 2 + strlen(*wptr));
      strcpy(next, lmatch);
      strcat(next, " ");
      strcat(next, *wptr);

      palindrome_worker(next, rmatch, (wl == n) ? NULL : (*wptr) + n, 1,
                        level + 1);
      free(next);
    }
    wptr = next_match(wptr, unmatch);
  }
}

main()
{
  char **wptr;

#ifdef RANDOM
  srand48((long int) time(NULL));
#endif

  read_words(&words, &nwords, LIBDIR "/words");
  read_words(&sdrow, &nsdrow, LIBDIR "/sdrow");

  wptr = words;

#ifdef RANDOM
  while (1) {
    wptr = random_lookup(words, nwords, "", 0);
    palindrome_worker(*wptr, "", *wptr, 1, 0);
  }
#else
  while (*wptr) {
    palindrome_worker(*wptr, "", *wptr, 1, 0);
    wptr++;
  }
#endif

  return 0;
}
